// $Id: SampleEventHandler.C,v 1.10 2006/05/11 22:39:37 emarkus Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

// A sample client for ORTS that shows how to access client data

// todo/questions:
//
// game scripts:
// - what to do with workers during build time
// - similarly, how to define mining?
// - how is attack range defined (distance between centers/hulls)?
//
// client:
// - how to access global player data?
// - how to compute the set of visible tiles?
// - how to compute tiles visible from a unit?
// - how to decide what can be attacked by a specific unit?

// FORS/FORALL/ERR macros are defined in orts3/libs/kernel/src/Global.H

#include "SampleEventHandler.H"
#include "GameStateModule.H"
//#include "GfxModule.H"
#include "Game.H"
#include "GameChanges.H"
#include "GameTile.H"
#include "ServerObjData.H"
#include "Options.H"

using namespace std;

// tile types from core.bp
// fixme: game script reflection would help here
// in the competition we won't use water tiles
// TILE_CLIFF: non-traversable by ground units

enum { TILE_UNKNOWN=0, TILE_WATER=1, TILE_GROUND=2, TILE_CLIFF=3 };

bool SampleEventHandler::handle_event(const Event& e)
{
	//	cout << "This is *creed's* event-handler" << endl;
	if (e.get_who() == GameStateModule::FROM) {

		//cout << "Event: " << e.who() << " " << e.what() << endl;

		if (e.get_what() == GameStateModule::STOP_MSG) {
			state.quit = true; return true;
		}

		if (e.get_what() == GameStateModule::READ_ERROR_MSG) {
			state.quit = state.error = true; return true;
		}

		if (e.get_what() == GameStateModule::VIEW_MSG) {

			compute_actions();
			state.gsm->send_actions();

			Game &game = state.gsm->get_game();	

			// gfx and frame statistics

			//      if (state.gfxm) state.gfxm->process_changes();

			sint4 vf = game.get_view_frame();
			sint4 af = game.get_action_frame();
			sint4 sa = game.get_skipped_actions();
			sint4 ma = game.get_merged_actions();

			if ((vf % 20) == 0) cout << "[frame " << vf << "]" << endl;

			if (vf != af || sa || ma > 1) {

				cout << "frame " << vf;
				if (af < 0) cout << " [no action]";
				else if (vf != af) cout << " [behind by " << (vf - af) << " frame(s)]";
				if (sa) cout << " [skipped " << sa << "]";
				if (ma > 1) cout << " [merged " << ma << "]";
				cout << endl;
			}

			/*      if (state.gfxm) {

			// don't draw if we are far behind
			if (abs(vf-af) < 10) {
			state.gfxm->draw();
			state.just_drew = true;
			}
			}*/

			return true;
		}
	}

	return false;
}

void SampleEventHandler::init_miner(GameObj* miner)
{
	if(assignments.find(miner)==assignments.end())
	{
		Job *job=new Job();
		job->target=NULL;
		job->laststopped=true; // to prevent first stop-check
		job->job=JOB_GOMINE;
		assignments[miner]=job;
	}
}

void SampleEventHandler::get_target(int &tx, int &ty, GameObj* miner)
{
	static set<GameObj*>::iterator i=resources.begin();

	bool wasistopped=false;
// initialize if isn't
	init_miner(miner);
	Job *job=assignments[miner];
// check if I was stopped
	if((*miner->sod.x!=job->gx || *miner->sod.x!=job->gy) && job->laststopped==false)
	{
		wasistopped=true;
		job->laststopped=true;
		cout << "Oh man! I was stopped!" << endl;
	}

	
	if(wasistopped)
	{
		ServerObjData &sod=miner->sod;
		int rx,ry; 
		real8 rdist=1000000000000000000000.; // big real ;)
		GameObj *res=NULL;
		// find nearest resource
		for (set<GameObj*>::iterator i = resources.begin(); i != resources.end();++i)
		{
			ServerObjData &sod2=(*i)->sod;

			real8 d2=(*sod.x-*sod2.x)*(*sod.x-*sod2.x)+
				(*sod.y-*sod2.y)*(*sod.y-*sod2.y);

			if(d2<rdist)
			{
				rdist=d2;
				rx=*sod2.x;
				ry=*sod2.y;
				res=*i;
			}
		}

		if(job->job==JOB_GOMINE && Game::distance(*miner,*res)<=2) // if we're looking for minerals and they're close enough
		{
			cout << "Jeee! I hit the minerals :) Gimme, gimme!" << endl;
			job->target=res; // little trick ;)
		}
		else
		{ // just pick any goal at random
#define LEAP_X 20
#define LEAP_Y 20
			tx=*miner->sod.x + (rand.rand_uint4()%(2*LEAP_X))-LEAP_X;
			ty=*miner->sod.y + (rand.rand_uint4()%(2*LEAP_Y))-LEAP_Y;
			return;
		}
	}
	
	if(job->target==NULL)
	{
		if(job->job==JOB_GOMINE)
		{
	        job->target=*i;
			i++; if(i==resources.end()) i=resources.begin();
		}
		else if(job->job==JOB_GOHOME)
		{
			job->target=base;			
		}
		else
		{
			cout << "I don't know what to do!!!";
			return;
		}
	}

	GameObj* mineral=job->target;
	tx=*mineral->sod.x;
	ty=*mineral->sod.y;

	job->laststopped=false;
	
	return;
}

real8 SampleEventHandler::distance2(GameObj* gob1, GameObj* gob2)
{
	ServerObjData &sod1=gob1->sod;
	ServerObjData &sod2=gob2->sod;

	real8 dx=*sod1.x-*sod2.x;
	real8 dy=*sod1.y-*sod2.y;
	return dx*dx+dy*dy;
}

void SampleEventHandler::compute_actions()
{
	cout << endl;

	cout << "compute actions" << endl;

	const GameChanges &changes = state.gsm->get_changes();

	// my code

	FORALL(changes.new_objs, i)
	{
		GameObj* gob=(*i)->get_GameObj();
		if(gob->bp_name() == "mineral")
			resources.insert(gob);
		else if(gob->bp_name() == "controlCenter")
			base=gob;
	}

	FORALL(changes.dead_objs, i)
	{
		GameObj *gob=(*i)->get_GameObj();
		if(gob->bp_name()=="mineral")
			resources.erase(gob);
		cout << "removed mineral" << endl;
	}

	const Game &game = state.gsm->get_game();
	const sint4 cid = game.get_client_player();
	const Game::ObjCont &objs = game.get_objs(cid);

	FORALL (objs, it) {
		GameObj *gob = (*it)->get_GameObj();
		if(*gob->sod.speed==0 && gob->sod.in_game && !gob->is_dead() && gob->bp_name()=="worker")
		{
			ScalarPoint t;

			init_miner(gob);
			GameObj *target=assignments[gob]->target;
			
			switch(assignments[gob]->job)
			{
				case JOB_MINE:
					if(gob->component("pickaxe")->get_int("active")!=0)
					{
						cout << "I'm still mining!" << endl;
					}
					else
					{
						cout << "Finished mining!" << endl;
						cout << "I have " << gob->get_int("minerals") << "minerals" << endl;
						assignments[gob]->job=JOB_GOHOME;
					}
					break;

				case JOB_GOMINE:
					if(target!=NULL && Game::distance(*target,*gob)<=2)
					{
						cout << "I'm near minerals! mining!" << endl;

						Vector<sint4> args;
						args.push_back(game.get_cplayer_info().get_id(target));
						gob->component("pickaxe")->set_action("mine",args);

						assignments[gob]->job=JOB_MINE;
						assignments[gob]->target=NULL;
					}
					else
					{
						get_target(t.x,t.y,gob);

						Vector<sint4> args;
						args.push_back(t.x);
						args.push_back(t.y);
						args.push_back(*gob->sod.max_speed);
	
						gob->set_action("move", args);
					}
					break;

				case JOB_GOHOME:
					if(target!=NULL && Game::distance(*target,*gob)<=3)
					{
						cout << "I'm near base! returning!" << endl;

						Vector<sint4> args;
						args.push_back(game.get_cplayer_info().get_id(target));
						gob->set_action("return_resources",args);						
		
						assignments[gob]->job=JOB_RETURN;
						assignments[gob]->target=NULL;
					}
					else
					{
						get_target(t.x,t.y,gob);

                        Vector<sint4> args;
                        args.push_back(t.x);
                        args.push_back(t.y);
                        args.push_back(*gob->sod.max_speed);

                        gob->set_action("move", args);
					}
					break;

				case JOB_RETURN:
					if(gob->get_int("minerals")==0)
					{
						cout << "Returing finished" << endl;

						assignments[gob]->job=JOB_GOMINE;
					}
					else
					{
						cout << "I'm still returning!" << endl;
					}
					break;

				default: cout << "I don't know what to do!!!" << endl;
			}
//			break; // just one!
		}
//		if(gob->bp_name()=="worker")
//			break; // and let the fun begin!
	}


	//FORALL (objs, it) {
	//	GameObj *gob = (*it)->get_GameObj();
	//	ServerObjData &sod=gob->sod;

	//	int rx,ry; 
	//	real8 rdist=1000000000000000000000.; // big real ;)
	//	// find nearest resource
	//	//	FORALL(resources,res)
	//	for (set<GameObj*>::iterator i = resources.begin(); i != resources.end();++i)
	//	{
	//		ServerObjData &sod2=(*i)->sod;

	//		real8 d2=(*sod.x-*sod2.x)*(*sod.x-*sod2.x)+
	//			(*sod.y-*sod2.y)*(*sod.y-*sod2.y);

	//		if(d2<rdist)
	//		{
	//			rdist=d2;
	//			rx=*sod2.x;
	//			ry=*sod2.y;
	//		}
	//	}
	//	t.x=rx; t.y=ry;

	//	Vector<sint4> params;
	//	params.push_back(t.x);
	//	params.push_back(t.y);
	//	params.push_back(*gob->sod.max_speed);

	//	gob->set_action("move", params);
	//}
	return;
	// ends here

	bool cinfo;
	Options::get("-cinfo", cinfo);

	if (cinfo) {

		//	  cout << "creed's cinfo";
		// - what changed in the last simulation tick

		if (!changes.new_tile_indexes.empty())
			cout << "#new tiles = " << changes.new_tile_indexes.size() << endl;

		if (!changes.new_boundaries.empty())
			cout << "#new boundaries = " << changes.new_boundaries.size() << endl;

		// fixme: what about removed tiles and removed boundaries?
		// currently tiles and boundaries are never removed

		if (!changes.new_objs.empty())
		{
			cout << "#new objects = " << changes.new_objs.size() << endl;

			FORALL(changes.new_objs, i)
			{
				GameObj* gob=(*i)->get_GameObj();
				ServerObjData &sod=gob->sod;

				if(gob->bp_name()=="mineral")
				{
					cout << "new mineral at " << *sod.x << "x" << *sod.y << endl;
				}		  
			}	  
		}

		if (!changes.changed_objs.empty())
			cout << "#changed objects = " << changes.changed_objs.size() << endl;

		if (!changes.vanished_objs.empty())
			cout << "#vanished objects = " << changes.vanished_objs.size() << endl;

		if (!changes.dead_objs.empty())
			cout << "#dead objects = " << changes.dead_objs.size() << endl;

	}

//	const Game &game = state.gsm->get_game();
//	const sint4 cid = game.get_client_player();

	// ***************** terrain ******************

	{
		const Map<GameTile> &map = game.get_map();

		bool minfo;
		Options::get("-minfo", minfo);

		if (minfo) {

			// - map dimensions

			cout << "Map width: = " << map.get_width()  << " tiles" << endl;
			cout << "Map height: " << map.get_height() << " tiles" << endl;  

			// height field scale: HEIGHT_MULT = one tile width up, constant for now
			cout << "height mult: " << GameConst::HEIGHT_MULT << endl; 

			// - new/updated tile indexes

			cout << "# of new/updated tiles " << changes.new_tile_indexes.size() << endl;
			cout << "Index(es): ";
			FORALL (changes.new_tile_indexes, it) cout << *it << " ";
			cout << endl;

			// - new boundaries

			const Vector<ScriptObj*> &boundaries = changes.new_boundaries;

			FORALL (boundaries, it) {

				const GameObj *g = (*it)->get_GameObj();
				if (g == 0) continue;

				cout << "new boundary (" << *g->sod.x1 << "," << *g->sod.y1 << ")-("
					<< *g->sod.x2 << "," << *g->sod.y2 << ")"
					<< endl;
			}

			// - all boundaries

			const Game::ObjCont &all_boundaries = game.get_boundaries();
			cout << "total # of boundaries = " << all_boundaries.size() << endl;
		}


		bool tinfo;
		Options::get("-tinfo", tinfo);

		if (tinfo) {

			// - all tiles

			FORS (y, map.get_height()) {
				FORS (x, map.get_width()) {
					const GameTile &tile = map(x,y);
					cout << "[" << x << "," << y << "]: ";

					sint4 hNW, hNE, hSE, hSW;
					Tile::Split split;

					// get corner heights and split type
					tile.get_topo(hNW, hNE, hSE, hSW, split);

					// tile types: see enum on top
					// split types: see orts3/libs/kernel/src/Tile.H
					// if split == Tile::NO_SPLIT => typeW is the tile type

					cout << "split:" << split 
						<< " type w:" << tile.get_typeW()
						<< " type e:" << tile.get_typeE()
						<< " heights:" << hNW << " " << hNE << " " << hSE << " " << hSW
						<< endl;
				}

				cout << endl;
			}

			cout << endl;
		}


		// - currently visible tiles

		// ... to be implemented
	}

	// ***************** objects ******************

	bool oinfo;
	Options::get("-oinfo", oinfo);

	if (oinfo) {

		// units under client player control

		const Game::ObjCont &objs = game.get_objs(cid);

		FORALL (objs, it) {

			const GameObj *gob = (*it)->get_GameObj();
			if (!gob) ERR("not a game object!?");

			const ServerObjData &sod = gob->sod;

			// - fundamental properties

			// blueprint name ~ type name
			cout << "bp: " << gob->bp_name();

			if (!sod.in_game) { cout << " no gameobj" << endl; continue; }

			if (*sod.x < 0) { cout << "not on playfield" << endl; continue; }

			// object on playfield

			// sod.shape and sod.zcat values defined in orts3/libs/kernel/src/Object.H

			cout << "zcat: " << *sod.zcat
				<< " max-speed: " << *sod.max_speed
				<< " speed: " << *sod.speed 
				<< " moving: " << *sod.is_moving 
				<< " sight: " << *sod.sight;

			if (*sod.shape == Object::CIRCLE) {
				cout << " CIRCLE loc: (" << *sod.x << "," << *sod.y << ")"
					<< " r: "<< *sod.radius;

			} else if (*sod.shape == Object::RECTANGLE) {
				cout << " RECT dim: (" << *sod.x1 << "," << *sod.y1 << ")-("
					<< *sod.x2 << "," << *sod.y2 << ")";
			} else {
				cout << " UNKNOWN SHAPE";
			}

			// - other object properties

			// heading: 0 = stopped, 1...GAME_CONST::HEADING_N
			const sintptr *hd = gob->get_const_int_ptr("heading");
			if (hd) cout << " hd: " << GameConst::angle_from_dir(*hd, GameConst::HEADING_N);

			// ... more examples

			cout << endl;
		}
	}

	// - what an object currently sees

	// ... to be implemented

	// - setting object actions

	const Map<GameTile> &map = game.get_map();
//	const Game::ObjCont &objs = game.get_objs(cid);
	const sint4 points_x = map.get_width()  * game.get_tile_points();
	const sint4 points_y = map.get_height() * game.get_tile_points();  

	FORALL (objs, it) {

		GameObj *gob = (*it)->get_GameObj();
		ServerObjData &sod=gob->sod;

		if (gob && gob->sod.in_game && *gob->sod.x >= 0 && !gob->is_dead()) {

			ScriptObj *weapon = gob->component("weapon");
			if (!weapon) continue;

			const sint4 n = game.get_player_num();
			GameObj *target = 0;

			if (n > 1) { // there are other players

				// todo: aim at weakest opponent unit within range

				sintptr *prange = weapon->get_int_ptr("max_ground_range");
				if (!prange) continue;

				const sint4 range2 = *prange * *prange;

				FORS (i, n) {

					if (i == cid) continue;

					// check for targets (could be done much faster)

					const Game::ObjCont &opp_objs = game.get_objs(i);

					FORALL (opp_objs, it) {

						GameObj *gob2 = (*it)->get_GameObj();
						if (gob2 && gob2->sod.in_game && *gob2->sod.x >= 0 && !gob2->is_dead()) {
							real8 d2 =
								square(real8(*gob->sod.x - *gob2->sod.x)) +
								square(real8(*gob->sod.y - *gob2->sod.y));

							if (d2 <= range2) {
								// target within range
								target = gob2;
							}
						}
					}
				}
			}

			if (target) {

				cout << "attack" << endl;

				// attack target

				// fixme: observe cooldown

				//cout << "launch " << sqrt(min_d) << " " << sqrt(attack_d) << endl;
				Vector<sint4> args;
				args.push_back(game.get_cplayer_info().get_id(target));
				gob->component("weapon")->set_action("attack", args);
				continue;
			}

			if (*gob->sod.speed == 0) {

				// object not dead and speed == 0 => move action

				// todo: attack weakest opponent in range

				ScalarPoint t;

				if ((rand.rand_uint4() % 100) < 33) {

					// random target on fringe (x or y unchanged)

					switch (rand.rand_uint4() & 3) {
	  case 0: t.x = *gob->sod.x; t.y = 0; break;
	  case 1: t.x = *gob->sod.x; t.y = points_y-1; break;
	  case 2: t.x = 0; t.y = *gob->sod.y; break;           
	  case 3: t.x = points_x-1; t.y = *gob->sod.y; break;
					}

				} else {

					// random target on fringe

					switch (rand.rand_uint4() & 3) {

	  case 0:
		  t.x = 0;
		  t.y = rand.rand_uint4() % points_y;
		  break;

	  case 1:
		  t.x = points_x - 1;
		  t.y = rand.rand_uint4() % points_y;
		  break;

	  case 2:
		  t.x = rand.rand_uint4() % points_x;
		  t.y = 0;
		  break;

	  case 3:
		  t.x = rand.rand_uint4() % points_x;
		  t.y = points_y - 1;
		  break;
					}
				}

				int rx,ry; 
				real8 rdist=1000000000000000000000.; // big real ;)
				// find nearest resource
				//	FORALL(resources,res)
				for (set<GameObj*>::iterator i = resources.begin(); i != resources.end();++i)
				{
					ServerObjData &sod2=(*i)->sod;

					real8 d2=(*sod.x-*sod2.x)*(*sod.x-*sod2.x)+
						(*sod.y-*sod2.y)*(*sod.y-*sod2.y);

					if(d2<rdist)
					{
						rdist=d2;
						rx=*sod2.x;
						ry=*sod2.y;
					}
				}
				t.x=rx; t.y=ry;

				Vector<sint4> params;
				params.push_back(t.x);
				params.push_back(t.y);
				params.push_back(*gob->sod.max_speed);

				gob->set_action("move", params);
			}
		}
	}

	// - global player/game variables

	// ... to be implemented
}
